import attr
from enum import IntEnum
from functools import partial
from ufoLib2.types import (
    Optional,
    List,
    Text,
    Integer,
    OptText,
    OptInteger,
    OptFloat,
    OptNumber,
    OptBool,
    OptIntList,
    OptNumList,
)
from ufoLib2.objects.misc import AttrDictMixin
from ufoLib2.objects.guideline import Guideline


__all__ = ["Info", "GaspRangeRecord", "NameRecord", "WidthClass"]


def _positive(instance, attribute, value):
    if value < 0:
        raise ValueError(
            "'{name}' must be at least 0 (got {value!r})".format(
                name=attribute.name, value=value
            )
        )


_optional_positive = attr.validators.optional(_positive)


# or maybe use IntFlag?
class GaspBehavior(IntEnum):
    GRIDFIT = 0
    DOGRAY = 1
    SYMMETRIC_SMOOTHING = 2
    SYMMETRIC_GRIDFIT = 3


@attr.s(slots=True)
class GaspRangeRecord(AttrDictMixin):
    rangeMaxPPEM = attr.ib(validator=_positive, type=Integer)
    rangeGaspBehavior = attr.ib(
        converter=lambda seq: [
            v if isinstance(v, GaspBehavior) else GaspBehavior(v) for v in seq
        ],
        type=List[GaspBehavior],  # use Set instead of List?
    )


@attr.s(slots=True)
class NameRecord(AttrDictMixin):
    nameID = attr.ib(validator=_positive, type=Integer)
    platformID = attr.ib(validator=_positive, type=Integer)
    encodingID = attr.ib(validator=_positive, type=Integer)
    languageID = attr.ib(validator=_positive, type=Integer)
    string = attr.ib(type=Text)


class WidthClass(IntEnum):
    ULTRA_CONDENSED = 1
    EXTRA_CONDESED = 2
    CONDENSED = 3
    SEMI_CONDENSED = 4
    NORMAL = 5  # alias for WidthClass.MEDIUM
    MEDIUM = 5
    SEMI_EXPANDED = 6
    EXPANDED = 7
    EXTRA_EXPANDED = 8
    ULTRA_EXPANDED = 9


def _convert_optional_list(lst, klass):
    if lst is None:
        return
    result = []
    for d in lst:
        if isinstance(d, klass):
            result.append(d)
        else:
            result.append(klass(**d))
    return result


_convert_guidelines = partial(_convert_optional_list, klass=Guideline)
_convert_gasp_range_records = partial(
    _convert_optional_list, klass=GaspRangeRecord
)
_convert_name_records = partial(_convert_optional_list, klass=NameRecord)


@attr.s(slots=True)
class Info(object):
    familyName = attr.ib(default=None, type=OptText)
    styleName = attr.ib(default=None, type=OptText)
    styleMapFamilyName = attr.ib(default=None, type=OptText)
    styleMapStyleName = attr.ib(default=None, type=OptText)
    versionMajor = attr.ib(
        default=None, validator=_optional_positive, type=OptInteger
    )
    versionMinor = attr.ib(
        default=None, validator=_optional_positive, type=OptInteger
    )

    copyright = attr.ib(default=None, type=OptText)
    trademark = attr.ib(default=None, type=OptText)

    unitsPerEm = attr.ib(
        default=None, validator=_optional_positive, type=OptNumber
    )
    descender = attr.ib(default=None, type=OptNumber)
    xHeight = attr.ib(default=None, type=OptNumber)
    capHeight = attr.ib(default=None, type=OptNumber)
    ascender = attr.ib(default=None, type=OptNumber)
    italicAngle = attr.ib(default=None, type=OptNumber)

    note = attr.ib(default=None, type=OptText)

    _guidelines = attr.ib(
        default=None,
        converter=_convert_guidelines,
        type=Optional[List[Guideline]],
    )

    @property
    def guidelines(self):
        return self._guidelines

    @guidelines.setter
    def guidelines(self, value):
        self._guidelines = _convert_guidelines(value)

    _openTypeGaspRangeRecords = attr.ib(
        default=None,
        converter=_convert_gasp_range_records,
        type=Optional[List[GaspRangeRecord]],
    )

    @property
    def openTypeGaspRangeRecords(self):
        return self._openTypeGaspRangeRecords

    @openTypeGaspRangeRecords.setter
    def openTypeGaspRangeRecords(self, value):
        self._openTypeGaspRangeRecords = _convert_gasp_range_records(value)

    openTypeHeadCreated = attr.ib(default=None, type=OptText)
    openTypeHeadLowestRecPPEM = attr.ib(
        default=None, validator=_optional_positive, type=OptInteger
    )
    openTypeHeadFlags = attr.ib(default=None, type=OptIntList)

    openTypeHheaAscender = attr.ib(default=None, type=OptInteger)
    openTypeHheaDescender = attr.ib(default=None, type=OptInteger)
    openTypeHheaLineGap = attr.ib(default=None, type=OptInteger)
    openTypeHheaCaretSlopeRise = attr.ib(default=None, type=OptInteger)
    openTypeHheaCaretSlopeRun = attr.ib(default=None, type=OptInteger)
    openTypeHheaCaretOffset = attr.ib(default=None, type=OptInteger)

    openTypeNameDesigner = attr.ib(default=None, type=OptText)
    openTypeNameDesignerURL = attr.ib(default=None, type=OptText)
    openTypeNameManufacturer = attr.ib(default=None, type=OptText)
    openTypeNameManufacturerURL = attr.ib(default=None, type=OptText)
    openTypeNameLicense = attr.ib(default=None, type=OptText)
    openTypeNameLicenseURL = attr.ib(default=None, type=OptText)
    openTypeNameVersion = attr.ib(default=None, type=OptText)
    openTypeNameUniqueID = attr.ib(default=None, type=OptText)
    openTypeNameDescription = attr.ib(default=None, type=OptText)
    openTypeNamePreferredFamilyName = attr.ib(default=None, type=OptText)
    openTypeNamePreferredSubfamilyName = attr.ib(default=None, type=OptText)
    openTypeNameCompatibleFullName = attr.ib(default=None, type=OptText)
    openTypeNameSampleText = attr.ib(default=None, type=OptText)
    openTypeNameWWSFamilyName = attr.ib(default=None, type=OptText)
    openTypeNameWWSSubfamilyName = attr.ib(default=None, type=OptText)

    _openTypeNameRecords = attr.ib(
        default=None,
        converter=_convert_name_records,
        type=Optional[List[NameRecord]],
    )

    @property
    def openTypeNameRecords(self):
        return self._openTypeNameRecords

    @openTypeNameRecords.setter
    def openTypeNameRecords(self, value):
        self._openTypeNameRecords = _convert_name_records(value)

    _openTypeOS2WidthClass = attr.ib(
        default=None,
        converter=lambda v: v if v is None else WidthClass(v),
        type=Optional[WidthClass],
    )

    @property
    def openTypeOS2WidthClass(self):
        return self._openTypeOS2WidthClass

    @openTypeOS2WidthClass.setter
    def openTypeOS2WidthClass(self, value):
        self._openTypeOS2WidthClass = (
            value if value is None else WidthClass(value)
        )

    openTypeOS2WeightClass = attr.ib(default=None, type=OptInteger)

    @openTypeOS2WeightClass.validator
    def _validate_weight_class(self, attribute, value):
        if value is not None and (value < 1 or value > 1000):
            raise ValueError(
                "'openTypeOS2WeightClass' must be between 1 and 1000"
            )

    openTypeOS2Selection = attr.ib(default=None, type=OptIntList)
    openTypeOS2VendorID = attr.ib(default=None, type=OptText)
    openTypeOS2Panose = attr.ib(default=None, type=OptIntList)
    openTypeOS2FamilyClass = attr.ib(default=None, type=OptIntList)
    openTypeOS2UnicodeRanges = attr.ib(default=None, type=OptIntList)
    openTypeOS2CodePageRanges = attr.ib(default=None, type=OptIntList)
    openTypeOS2TypoAscender = attr.ib(default=None, type=OptInteger)
    openTypeOS2TypoDescender = attr.ib(default=None, type=OptInteger)
    openTypeOS2TypoLineGap = attr.ib(default=None, type=OptInteger)
    openTypeOS2WinAscent = attr.ib(
        default=None, validator=_optional_positive, type=OptInteger
    )
    openTypeOS2WinDescent = attr.ib(
        default=None, validator=_optional_positive, type=OptInteger
    )
    openTypeOS2Type = attr.ib(default=None, type=OptIntList)
    openTypeOS2SubscriptXSize = attr.ib(default=None, type=OptInteger)
    openTypeOS2SubscriptYSize = attr.ib(default=None, type=OptInteger)
    openTypeOS2SubscriptXOffset = attr.ib(default=None, type=OptInteger)
    openTypeOS2SubscriptYOffset = attr.ib(default=None, type=OptInteger)
    openTypeOS2SuperscriptXSize = attr.ib(default=None, type=OptInteger)
    openTypeOS2SuperscriptYSize = attr.ib(default=None, type=OptInteger)
    openTypeOS2SuperscriptXOffset = attr.ib(default=None, type=OptInteger)
    openTypeOS2SuperscriptYOffset = attr.ib(default=None, type=OptInteger)
    openTypeOS2StrikeoutSize = attr.ib(default=None, type=OptInteger)
    openTypeOS2StrikeoutPosition = attr.ib(default=None, type=OptInteger)

    openTypeVheaVertTypoAscender = attr.ib(default=None, type=OptInteger)
    openTypeVheaVertTypoDescender = attr.ib(default=None, type=OptInteger)
    openTypeVheaVertTypoLineGap = attr.ib(default=None, type=OptInteger)
    openTypeVheaCaretSlopeRise = attr.ib(default=None, type=OptInteger)
    openTypeVheaCaretSlopeRun = attr.ib(default=None, type=OptInteger)
    openTypeVheaCaretOffset = attr.ib(default=None, type=OptInteger)

    postscriptFontName = attr.ib(default=None, type=OptText)
    postscriptFullName = attr.ib(default=None, type=OptText)
    postscriptSlantAngle = attr.ib(default=None, type=OptNumber)
    postscriptUniqueID = attr.ib(default=None, type=OptInteger)
    postscriptUnderlineThickness = attr.ib(default=None, type=OptNumber)
    postscriptUnderlinePosition = attr.ib(default=None, type=OptNumber)
    postscriptIsFixedPitch = attr.ib(default=None, type=OptBool)
    postscriptBlueValues = attr.ib(default=None, type=OptNumList)
    postscriptOtherBlues = attr.ib(default=None, type=OptNumList)
    postscriptFamilyBlues = attr.ib(default=None, type=OptNumList)
    postscriptFamilyOtherBlues = attr.ib(default=None, type=OptNumList)
    postscriptStemSnapH = attr.ib(default=None, type=OptNumList)
    postscriptStemSnapV = attr.ib(default=None, type=OptNumList)
    postscriptBlueFuzz = attr.ib(default=None, type=OptNumber)
    postscriptBlueShift = attr.ib(default=None, type=OptNumber)
    postscriptBlueScale = attr.ib(default=None, type=OptFloat)
    postscriptForceBold = attr.ib(default=None, type=OptBool)
    postscriptDefaultWidthX = attr.ib(default=None, type=OptNumber)
    postscriptNominalWidthX = attr.ib(default=None, type=OptNumber)
    postscriptWeightName = attr.ib(default=None, type=OptText)
    postscriptDefaultCharacter = attr.ib(default=None, type=OptText)
    postscriptWindowsCharacterSet = attr.ib(default=None, type=OptText)

    # old stuff
    macintoshFONDName = attr.ib(default=None, type=OptText)
    macintoshFONDFamilyID = attr.ib(default=None, type=OptInteger)
    year = attr.ib(default=None, type=OptInteger)



import py
import pytest
import ufoLib2


@pytest.fixture
def datadir(request):
    return py.path.local(request.fspath.dirname).join("data")


@pytest.fixture
def ufo_UbuTestData(tmpdir, datadir):
    ufo_path = tmpdir.join("UbuTestData.ufo")
    datadir.join("UbuTestData.ufo").copy(ufo_path)
    return ufoLib2.Font.open(str(ufo_path))



from __future__ import unicode_literals

DATA_DIRNAME = "data"
DEFAULT_GLYPHS_DIRNAME = "glyphs"
IMAGES_DIRNAME = "images"

FEATURES_FILENAME = "features.fea"
FONTINFO_FILENAME = "fontinfo.plist"
GROUPS_FILENAME = "groups.plist"
KERNING_FILENAME = "kerning.plist"
LAYERCONTENTS_FILENAME = "layercontents.plist"
LIB_FILENAME = "lib.plist"
METAINFO_FILENAME = "metainfo.plist"

CONTENTS_FILENAME = "contents.plist"
LAYERINFO_FILENAME = "layerinfo.plist"

DEFAULT_LAYER_NAME = "public.default"



from fontTools.pens.pointPen import AbstractPointPen
from ufoLib2.objects.component import Component
from ufoLib2.objects.contour import Contour
from ufoLib2.objects.point import Point


class GlyphPointPen(AbstractPointPen):
    __slots__ = "_glyph", "_contour"

    def __init__(self, glyph):
        self._glyph = glyph
        self._contour = None

    def beginPath(self, identifier=None, **kwargs):
        self._contour = Contour(identifier=identifier)

    def endPath(self):
        self._glyph.contours.append(self._contour)
        self._contour = None

    def addPoint(
        self,
        pt,
        segmentType=None,
        smooth=False,
        name=None,
        identifier=None,
        **kwargs
    ):
        x, y = pt
        self._contour.append(
            Point(
                x,
                y,
                type=segmentType,
                smooth=smooth,
                name=name,
                identifier=identifier,
            )
        )

    def addComponent(self, baseGlyph, transformation, **kwargs):
        component = Component(baseGlyph, transformation, **kwargs)
        self._glyph.components.append(component)



from __future__ import absolute_import, unicode_literals
import attr
from fontTools.misc.py23 import tounicode, unicode


@attr.s(slots=True)
class Features(object):
    text = attr.ib(default="", converter=tounicode, type=unicode)

    def __bool__(self):
        return bool(self.text)

    __nonzero__ = __bool__

    def __str__(self):
        return self.text



from ufoLib2.objects.misc import DataStore
from fontTools.ufoLib import UFOReader
from fontTools.ufoLib import UFOWriter


class ImageSet(DataStore):
    listdir = UFOReader.getImageDirectoryListing
    readf = UFOReader.readImage
    writef = UFOWriter.writeImage
    deletef = UFOWriter.removeImage



import attr
from typing import Optional, Union
from ufoLib2.objects.misc import AttrDictMixin


@attr.s(slots=True)
class Anchor(AttrDictMixin):
    x = attr.ib(type=Union[int, float])
    y = attr.ib(type=Union[int, float])
    name = attr.ib(default=None, type=Optional[str])
    color = attr.ib(default=None, type=Optional[str])
    identifier = attr.ib(default=None, type=Optional[str])



import attr
from typing import Optional
from ufoLib2.objects.glyph import Glyph
from ufoLib2.objects.misc import _NOT_LOADED
from ufoLib2.constants import DEFAULT_LAYER_NAME


def _convert_glyphs(value):
    if isinstance(value, dict):
        return value
    result = {}
    for glyph in value:
        if glyph.name in result:
            raise KeyError("glyph %r already exists" % glyph.name)
        result[glyph.name] = glyph
    return result


@attr.s(slots=True, repr=False)
class Layer(object):
    _name = attr.ib(default=DEFAULT_LAYER_NAME, type=str)
    _glyphs = attr.ib(
        default=attr.Factory(dict), converter=_convert_glyphs, type=dict
    )
    color = attr.ib(default=None, type=Optional[str])
    lib = attr.ib(default=attr.Factory(dict), type=dict)

    _glyphSet = attr.ib(default=None, init=False)

    @classmethod
    def read(cls, name, glyphSet, lazy=True):
        glyphNames = glyphSet.keys()
        if lazy:
            glyphs = {gn: _NOT_LOADED for gn in glyphNames}
        else:
            glyphs = {}
            for glyphName in glyphNames:
                glyph = Glyph(glyphName)
                glyphSet.readGlyph(glyphName, glyph, glyph.getPointPen())
                glyphs[glyphName] = glyph
        self = cls(name, glyphs)
        if lazy:
            self._glyphSet = glyphSet
        glyphSet.readLayerInfo(self)
        return self

    def __contains__(self, name):
        return name in self._glyphs

    def __delitem__(self, name):
        del self._glyphs[name]

    def __getitem__(self, name):
        if self._glyphs[name] is _NOT_LOADED:
            return self.loadGlyph(name)
        return self._glyphs[name]

    def __setitem__(self, name, glyph):
        if not isinstance(glyph, Glyph):
            raise TypeError("Expected Glyph, found %s" % type(glyph).__name__)
        glyph._name = name
        self._glyphs[name] = glyph

    def __iter__(self):
        for name in self._glyphs:
            yield self[name]

    def __len__(self):
        return len(self._glyphs)

    def __repr__(self):
        n = len(self._glyphs)
        return "<{}.{} '{}' ({}) at {}>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            self._name,
            "empty"
            if n == 0
            else "{} glyph{}".format(n, "s" if n > 1 else ""),
            hex(id(self)),
        )

    def get(self, name, default=None):
        try:
            return self[name]
        except KeyError:
            return default

    def keys(self):
        return self._glyphs.keys()

    def pop(self, name, default=KeyError):
        try:
            glyph = self[name]
        except KeyError:
            if default is KeyError:
                raise
            glyph = default
        else:
            del self[name]
        return glyph

    @property
    def name(self):
        return self._name

    def addGlyph(self, glyph):
        if glyph.name in self._glyphs:
            raise KeyError("glyph %r already exists" % glyph.name)
        self._glyphs[glyph.name] = glyph

    def loadGlyph(self, name):
        glyph = Glyph(name)
        self._glyphSet.readGlyph(name, glyph, glyph.getPointPen())
        self._glyphs[name] = glyph
        return glyph

    def newGlyph(self, name):
        if name in self._glyphs:
            raise KeyError("glyph %r already exists" % name)
        self._glyphs[name] = glyph = Glyph(name)
        return glyph

    def renameGlyph(self, name, newName, overwrite=False):
        if name == newName:
            return
        if not overwrite and newName in self._glyphs:
            raise KeyError("target glyph %r already exists" % newName)
        # pop and set name
        glyph = self.pop(name)
        glyph._name = newName
        # add it back
        self._glyphs[newName] = glyph

    def write(self, glyphSet, saveAs=True):
        glyphs = self._glyphs
        if not saveAs:
            for name in set(glyphSet.contents).difference(glyphs):
                glyphSet.deleteGlyph(name)
        for name, glyph in glyphs.items():
            if glyph is _NOT_LOADED:
                if saveAs:
                    glyph = self.loadGlyph(name)
                else:
                    continue
            glyphSet.writeGlyph(
                name, glyphObject=glyph, drawPointsFunc=glyph.drawPoints
            )
        glyphSet.writeContents()
        glyphSet.writeLayerInfo(self)
        if saveAs:
            # all glyphs are loaded by now, no need to keep ref to glyphSet
            self._glyphSet = None



import attr
from typing import Optional, Union
from ufoLib2.objects.misc import AttrDictMixin


@attr.s(slots=True)
class Guideline(AttrDictMixin):
    x = attr.ib(default=None, type=Optional[Union[int, float]])
    y = attr.ib(default=None, type=Optional[Union[int, float]])
    angle = attr.ib(default=None, type=Optional[Union[int, float]])
    name = attr.ib(default=None, type=Optional[str])
    color = attr.ib(default=None, type=Optional[str])
    identifier = attr.ib(default=None, type=Optional[str])

    def __attrs_post_init__(self):
        x, y, angle = self.x, self.y, self.angle
        if x is None and y is None:
            raise ValueError("x or y must be present")
        if x is None or y is None:
            if angle is not None:
                raise ValueError(
                    "if 'x' or 'y' are None, 'angle' must not be present"
                )
        if x is not None and y is not None and angle is None:
            raise ValueError(
                "if 'x' and 'y' are defined, 'angle' must be defined"
            )
        if angle is not None and not (0 <= angle <= 360):
            raise ValueError("angle must be between 0 and 360")



import attr

try:
    from collections.abc import Mapping, MutableMapping
except ImportError:
    from collections import Mapping, MutableMapping


# sentinel value to signal a "lazy" object hasn't been loaded yet
_NOT_LOADED = object()


@attr.s(slots=True, repr=False)
class DataStore(MutableMapping):
    listdir = None
    readf = None
    writef = None
    deletef = None

    _data = attr.ib(default=attr.Factory(dict), type=dict)

    _reader = attr.ib(default=None, init=False, repr=False)
    _scheduledForDeletion = attr.ib(
        default=attr.Factory(set), init=False, repr=False
    )

    @classmethod
    def read(cls, reader, lazy=True):
        self = cls()
        for fileName in cls.listdir(reader):
            if lazy:
                self._data[fileName] = _NOT_LOADED
            else:
                self._data[fileName] = cls.readf(reader, fileName)
        if lazy:
            self._reader = reader
        return self

    # MutableMapping methods

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __getitem__(self, fileName):
        if self._data[fileName] is _NOT_LOADED:
            self._data[fileName] = self.__class__.readf(self._reader, fileName)
        return self._data[fileName]

    def __setitem__(self, fileName, data):
        # should we forbid overwrite?
        self._data[fileName] = data
        if fileName in self._scheduledForDeletion:
            self._scheduledForDeletion.remove(fileName)

    def __delitem__(self, fileName):
        del self._data[fileName]
        self._scheduledForDeletion.add(fileName)

    def __repr__(self):
        n = len(self._data)
        return "<{}.{} ({}) at {}>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            "empty" if n == 0 else "{} file{}".format(n, "s" if n > 1 else ""),
            hex(id(self)),
        )

    def write(self, writer, saveAs=None):
        if saveAs is None:
            saveAs = self._reader is not writer
        # if in-place, remove deleted data
        if not saveAs:
            for fileName in self._scheduledForDeletion:
                self.__class__.deletef(writer, fileName)
        # Write data. Iterating over _data.items() prevents automatic loading.
        for fileName, data in self._data.items():
            # Two paths:
            # 1) We are saving in-place. Only write to disk what is loaded, it
            #    might be modified.
            # 2) We save elsewhere. Load all data files to write them back out.
            if data is _NOT_LOADED:
                if saveAs:
                    data = self.__class__.readf(self._reader, fileName)
                    self._data[fileName] = data
                else:
                    continue
            self.__class__.writef(writer, fileName, data)
        self._scheduledForDeletion = set()
        if saveAs:
            # all data was read by now, ref to reader no longer needed
            self._reader = None

    @property
    def fileNames(self):
        return list(self._data.keys())


class AttrDictMixin(Mapping):
    """ Read attribute values using mapping interface. For use with Anchors and
    Guidelines classes, where client code expects them to behave as dict.
    """

    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(key)

    def __iter__(self):
        for key in attr.fields_dict(self.__class__):
            if getattr(self, key) is not None:
                yield key

    def __len__(self):
        return sum(1 for _ in self)



import ufoLib2
from fontTools.misc.py23 import basestring
from ufoLib2.objects.misc import _NOT_LOADED


def test_import_version():
    assert hasattr(ufoLib2, "__version__")
    assert isinstance(ufoLib2.__version__, basestring)


def test_LayerSet_load_layers_on_iteration(tmpdir):
    ufo = ufoLib2.Font()
    ufo.layers.newLayer("test")
    ufo_save_path = str(tmpdir.join("test.ufo"))
    ufo.save(ufo_save_path)
    ufo = ufoLib2.Font.open(ufo_save_path)
    for layer in ufo.layers:
        assert layer is not _NOT_LOADED


def test_lazy_data_loading_saveas(ufo_UbuTestData, tmpdir):
    ufo = ufo_UbuTestData
    ufo_path = str(tmpdir.join("UbuTestData2.ufo"))
    ufo.save(ufo_path)
    assert all(v is not _NOT_LOADED for v in ufo.data._data.values())


def test_lazy_data_loading_inplace_no_load(ufo_UbuTestData):
    ufo = ufo_UbuTestData
    ufo.save()
    assert all(v is _NOT_LOADED for v in ufo.data._data.values())


def test_lazy_data_loading_inplace_load_some(ufo_UbuTestData):
    ufo = ufo_UbuTestData
    some_data = "abc".encode("ascii")
    ufo.data["com.github.fonttools.ttx/T_S_I__0.ttx"] = some_data
    ufo.save()
    assert all(
        v is _NOT_LOADED
        for k, v in ufo.data._data.items()
        if "T_S_I__0" not in k
    )
    assert ufo.data["com.github.fonttools.ttx/T_S_I__0.ttx"] == some_data



import attr
from typing import Optional
from fontTools.misc.transform import Identity, Transform

try:
    from collections.abc import Mapping  # python >= 3.3
except ImportError:
    from collections import Mapping


@attr.s(slots=True)
class Image(Mapping):
    fileName = attr.ib(default=None, type=Optional[str])
    transformation = attr.ib(
        default=Identity,
        convert=lambda t: t if isinstance(t, Transform) else Transform(*t),
        type=Transform,
    )
    color = attr.ib(default=None, type=Optional[str])

    def clear(self):
        self.fileName = None
        self.transformation = Identity
        self.color = None

    def __bool__(self):
        # Glyph.image evaluates to False if no fileName is set
        return self.fileName is not None

    # alias for python 2
    __nonzero__ = __bool__

    _transformation_keys_ = (
        "xScale",
        "xyScale",
        "yxScale",
        "yScale",
        "xOffset",
        "yOffset",
    )
    _valid_keys_ = ("fileName",) + _transformation_keys_ + ("color",)

    # implementation of collections.abc.Mapping abstract methods.
    # the fontTools.ufoLib.validators.imageValidator requires that image is a
    # subclass of Mapping...

    def __getitem__(self, key):
        try:
            i = self._transformation_keys_.index(key)
        except ValueError:
            try:
                return getattr(self, key)
            except AttributeError:
                raise KeyError(key)
        else:
            return self.transformation[i]

    def __len__(self):
        return len(self._valid_keys_)

    def __iter__(self):
        return iter(self._valid_keys_)



import attr
from typing import Optional, Union


@attr.s(slots=True)
class Point(object):
    x = attr.ib(type=Union[int, float])
    y = attr.ib(type=Union[int, float])
    type = attr.ib(type=Optional[str])
    smooth = attr.ib(default=False, type=bool)
    name = attr.ib(default=None, type=Optional[str])
    identifier = attr.ib(default=None, type=Optional[str])

    @property
    def segmentType(self):
        # alias for backward compatibility with defcon API
        return self.type



from typing import Optional, Union, Text, List
from fontTools.misc.py23 import PY3


try:
    long
except NameError:
    long = int


OptText = Optional[Text]
Integer = int if PY3 else Union[int, long]
Number = Union[float, Integer]
OptInteger = Optional[Integer]
OptFloat = Optional[float]
OptNumber = Optional[Number]
OptBool = Optional[bool]

IntList = List[Integer]
OptIntList = Optional[IntList]
NumList = List[Number]
OptNumList = Optional[NumList]



import attr
from collections import OrderedDict
from ufoLib2.objects.misc import _NOT_LOADED
from ufoLib2.objects.layer import Layer
from ufoLib2.constants import DEFAULT_LAYER_NAME


def _convert_layers(value):
    # takes an iterable of Layer objects and returns an OrderedDict keyed
    # by layer name
    if isinstance(value, OrderedDict):
        return value
    layers = OrderedDict()
    for layer in value:
        if not isinstance(layer, Layer):
            raise TypeError(
                "expected 'Layer', found '%s'" % type(layer).__name__
            )
        if layer.name in layers:
            raise KeyError("duplicate layer name: '%s'" % layer.name)
        layers[layer.name] = layer
    return layers


@attr.s(slots=True, repr=False)
class LayerSet(object):
    _layers = attr.ib(default=(), converter=_convert_layers, type=OrderedDict)
    defaultLayer = attr.ib(default=None, type=Layer)

    _reader = attr.ib(default=None, init=False)

    def __attrs_post_init__(self):
        if not self._layers:
            # LayerSet is never empty; always contains at least the default
            if self.defaultLayer is not None:
                raise TypeError(
                    "'defaultLayer' argument is invalid with empty LayerSet"
                )
            self.defaultLayer = self.newLayer(DEFAULT_LAYER_NAME)
        elif self.defaultLayer is not None:
            # check that the specified default layer is in the layer set;
            default = self.defaultLayer
            for layer in self._layers.values():
                if layer is default:
                    break
            else:
                raise ValueError(
                    "defaultLayer %r is not among the specified layers"
                    % default
                )
        elif len(self._layers) == 1:
            # there's only one, we assume it's the default
            self.defaultLayer = next(self._layers.values())
        else:
            if DEFAULT_LAYER_NAME not in self._layers:
                raise ValueError("default layer not specified")
            self.defaultLayer = self._layers[DEFAULT_LAYER_NAME]

    @classmethod
    def read(cls, reader, lazy=True):
        layers = OrderedDict()
        defaultLayer = None

        defaultLayerName = reader.getDefaultLayerName()

        for layerName in reader.getLayerNames():
            isDefault = layerName == defaultLayerName
            if isDefault or not lazy:
                layer = cls._loadLayer(reader, layerName, lazy, isDefault)
                if isDefault:
                    defaultLayer = layer
                layers[layerName] = layer
            else:
                layers[layerName] = _NOT_LOADED

        assert defaultLayer is not None

        self = cls(layers, defaultLayer=defaultLayer)
        if lazy:
            self._reader = reader

        return self

    @staticmethod
    def _loadLayer(reader, layerName, lazy=True, default=False):
        # UFOReader.getGlyphSet method doesn't support 'defaultLayer'
        # keyword argument, whereas the UFOWriter.getGlyphSet method
        # requires it. In order to allow to use an instance of
        # UFOWriter as the 'reader', here we try both ways...
        try:
            glyphSet = reader.getGlyphSet(layerName, defaultLayer=default)
        except TypeError as e:
            # TODO use inspect module?
            if "keyword argument 'defaultLayer'" in str(e):
                glyphSet = reader.getGlyphSet(layerName)

        return Layer.read(layerName, glyphSet, lazy=lazy)

    def loadLayer(self, layerName, lazy=True, default=False):
        assert self._reader is not None
        if layerName not in self._layers:
            raise KeyError(layerName)
        layer = self._loadLayer(self._reader, layerName, lazy, default)
        self._layers[layerName] = layer
        return layer

    def __contains__(self, name):
        return name in self._layers

    def __delitem__(self, name):
        if name == self.defaultLayer.name:
            raise KeyError("cannot delete default layer %r" % name)
        del self._layers[name]

    def __getitem__(self, name):
        if self._layers[name] is _NOT_LOADED:
            return self.loadLayer(name)
        return self._layers[name]

    def __iter__(self):
        for layer_name, layer_object in self._layers.items():
            if layer_object is _NOT_LOADED:
                yield self.loadLayer(layer_name)
            else:
                yield layer_object

    def __len__(self):
        return len(self._layers)

    def get(self, name, default=None):
        return self._layers.get(name, default)

    def keys(self):
        return self._layers.keys()

    def __repr__(self):
        n = len(self._layers)
        return "<{}.{} ({} layer{}) at {}>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            n,
            "s" if n > 1 else "",
            hex(id(self)),
        )

    @property
    def layerOrder(self):
        return list(self._layers)

    @layerOrder.setter
    def layerOrder(self, order):
        assert set(order) == set(self._layers)
        layers = OrderedDict()
        for name in order:
            layers[name] = self._layers[name]
        self._layers = layers

    def newLayer(self, name, **kwargs):
        if name in self._layers:
            raise KeyError("layer %r already exists" % name)
        self._layers[name] = layer = Layer(name, **kwargs)
        return layer

    def renameGlyph(self, name, newName, overwrite=False):
        # Note: this would be easier if the glyph contained the layers!
        if name == newName:
            return
        # make sure we're copying something
        if not any(name in layer for layer in self):
            raise KeyError("name %r is not in layer set" % name)
        # prepare destination, delete if overwrite=True or error
        for layer in self:
            if newName in self._layers:
                if overwrite:
                    del layer[newName]
                else:
                    raise KeyError("target name %r already exists" % newName)
        # now do the move
        for layer in self:
            if name in layer:
                layer[newName] = glyph = layer.pop(name)
                glyph._name = newName

    def renameLayer(self, name, newName, overwrite=False):
        if name == newName:
            return
        if not overwrite and newName in self._layers:
            raise KeyError("target name %r already exists" % newName)
        self._layers[newName] = layer = self._layers.pop(name)
        layer._name = newName

    def write(self, writer, saveAs=None):
        if saveAs is None:
            saveAs = self._reader is not writer
        # if in-place, remove deleted layers
        layers = self._layers
        if not saveAs:
            for name in set(writer.getLayerNames()).difference(layers):
                writer.deleteGlyphSet(name)
        # write layers
        defaultLayer = self.defaultLayer
        for name, layer in layers.items():
            default = layer is defaultLayer
            if layer is _NOT_LOADED:
                if saveAs:
                    layer = self.loadLayer(name, lazy=False)
                else:
                    continue
            glyphSet = writer.getGlyphSet(name, defaultLayer=default)
            layer.write(glyphSet, saveAs=saveAs)
        writer.writeLayerContents(self.layerOrder)



import attr
from typing import Optional
from fontTools.misc.transform import Identity, Transform
from fontTools.pens.pointPen import PointToSegmentPen
import warnings


@attr.s(slots=True)
class Component(object):
    baseGlyph = attr.ib(type=str)
    transformation = attr.ib(
        default=Identity,
        convert=lambda t: t if isinstance(t, Transform) else Transform(*t),
        type=Transform,
    )
    identifier = attr.ib(default=None, type=Optional[str])

    # -----------
    # Pen methods
    # -----------

    def draw(self, pen):
        pointPen = PointToSegmentPen(pen)
        self.drawPoints(pointPen)

    def drawPoints(self, pointPen):
        try:
            pointPen.addComponent(
                self.baseGlyph, self.transformation, identifier=self.identifier
            )
        except TypeError:
            pointPen.addComponent(self.baseGlyph, self.transformation)
            warnings.warn(
                "The addComponent method needs an identifier kwarg. "
                "The component's identifier value has been discarded.",
                UserWarning,
            )



import attr
from typing import Optional, Union, List, Any, Dict
from ufoLib2.objects.anchor import Anchor
from ufoLib2.objects.guideline import Guideline
from ufoLib2.objects.image import Image
from fontTools.pens.pointPen import PointToSegmentPen, SegmentToPointPen
from fontTools.misc.transform import Transform
from fontTools.misc.py23 import tounicode, unicode
from ufoLib2.pointPens.glyphPointPen import GlyphPointPen


@attr.s(slots=True, repr=False)
class Glyph(object):
    _name = attr.ib(type=str)
    width = attr.ib(default=0, type=Union[int, float])
    height = attr.ib(default=0, type=Union[int, float])
    unicodes = attr.ib(default=attr.Factory(list), type=List[int])

    _image = attr.ib(default=attr.Factory(Image), type=Image)
    lib = attr.ib(default=attr.Factory(dict), type=Dict[str, Any])
    note = attr.ib(
        default=None,
        converter=lambda n: tounicode(n) if n is not None else None,
        type=Optional[unicode],
    )
    _anchors = attr.ib(default=attr.Factory(list), type=list)
    components = attr.ib(default=attr.Factory(list), type=list)
    contours = attr.ib(default=attr.Factory(list), type=list)
    _guidelines = attr.ib(default=attr.Factory(list), type=list)

    def __len__(self):
        return len(self.contours)

    def __getitem__(self, index):
        return self.contours[index]

    def __contains__(self, contour):
        return contour in self.contours

    def __iter__(self):
        return iter(self.contours)

    def __repr__(self):
        return "<{}.{} '{}' at {}>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            self._name,
            hex(id(self)),
        )

    @property
    def anchors(self):
        return self._anchors

    @anchors.setter
    def anchors(self, value):
        self.clearAnchors()
        for anchor in value:
            self.appendAnchor(anchor)

    @property
    def guidelines(self):
        return self._guidelines

    @guidelines.setter
    def guidelines(self, value):
        self.clearGuidelines()
        for guideline in value:
            self.appendGuideline(guideline)

    @property
    def name(self):
        return self._name

    @property
    def unicode(self):
        if self.unicodes:
            return self.unicodes[0]
        return None

    @unicode.setter
    def unicode(self, value):
        if value is None:
            self.unicodes = []
        else:
            if self.unicodes:
                if self.unicodes[0] == value:
                    return
                try:
                    self.unicodes.remove(value)
                except ValueError:
                    pass
                self.unicodes.insert(0, value)
            else:
                self.unicodes.append(value)

    @property
    def image(self):
        return self._image

    @image.setter
    def image(self, image):
        if isinstance(image, Image):
            self._image = image
        else:
            self._image = Image(
                fileName=image["fileName"],
                transformation=Transform(
                    image["xScale"],
                    image["xyScale"],
                    image["yxScale"],
                    image["yScale"],
                    image["xOffset"],
                    image["yOffset"],
                ),
                color=image.get("color"),
            )

    def clear(self):
        del self._anchors[:]
        del self.components[:]
        del self.contours[:]
        del self._guidelines[:]
        self.image.clear()

    def clearAnchors(self):
        del self._anchors[:]

    def clearContours(self):
        del self.contours[:]

    def clearComponents(self):
        del self.components[:]

    def clearGuidelines(self):
        del self._guidelines[:]

    def removeComponent(self, component):
        self.components.remove(component)

    def appendAnchor(self, anchor):
        if not isinstance(anchor, Anchor):
            anchor = Anchor(**anchor)
        self.anchors.append(anchor)

    def appendGuideline(self, guideline):
        if not isinstance(guideline, Guideline):
            guideline = Guideline(**guideline)
        self._guidelines.append(guideline)

    # -----------
    # Pen methods
    # -----------

    def draw(self, pen):
        pointPen = PointToSegmentPen(pen)
        self.drawPoints(pointPen)

    def drawPoints(self, pointPen):
        for contour in self.contours:
            contour.drawPoints(pointPen)
        for component in self.components:
            component.drawPoints(pointPen)

    def getPen(self):
        pen = SegmentToPointPen(self.getPointPen())
        return pen

    def getPointPen(self):
        pointPen = GlyphPointPen(self)
        return pointPen

    # lib wrapped attributes

    @property
    def markColor(self):
        return self.lib.get("public.markColor")

    @markColor.setter
    def markColor(self, value):
        if value is not None:
            self.lib["public.markColor"] = value
        elif "public.markColor" in self.lib:
            del self.lib["public.markColor"]

    @property
    def verticalOrigin(self):
        return self.lib.get("public.verticalOrigin")

    @verticalOrigin.setter
    def verticalOrigin(self, value):
        if value is not None:
            self.lib["public.verticalOrigin"] = value
        elif "public.verticalOrigin" in self.lib:
            del self.lib["public.verticalOrigin"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
from io import open


with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="ufoLib2",
    use_scm_version={"write_to": "src/ufoLib2/_version.py"},
    description="ufoLib2 is a UFO font library.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Adrien Tétar",
    author_email="adri-from-59@hotmail.fr",
    url="https://github.com/adrientetar/ufoLib2",
    license="Apache 2.0",
    package_dir={"": "src"},
    packages=find_packages("src"),
    include_package_data=True,
    python_requires=">=2.7",
    install_requires=[
        "fonttools[ufo,lxml] >= 3.34.0",
        "attrs >= 18.2.0",
        "typing ; python_version < '3.5'",
        "enum34 ; python_version < '3.4'",
    ],
    setup_requires=["setuptools_scm"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Topic :: Text Processing :: Fonts",
        "License :: OSI Approved :: Apache Software License",
    ],
)



from ufoLib2.objects.misc import DataStore
from fontTools.ufoLib import UFOReader
from fontTools.ufoLib import UFOWriter


class DataSet(DataStore):
    listdir = UFOReader.getDataDirectoryListing
    readf = UFOReader.readData
    writef = UFOWriter.writeData
    deletef = UFOWriter.removeData



import attr
from typing import Optional

try:
    from collections.abc import MutableSequence
except ImportError:
    from collections import MutableSequence

import warnings
from fontTools.pens.pointPen import PointToSegmentPen

from ufoLib2.objects.point import Point


@attr.s(slots=True)
class Contour(MutableSequence):
    points = attr.ib(default=attr.Factory(list), type=list)
    identifier = attr.ib(default=None, repr=False, type=Optional[str])

    # collections.abc.MutableSequence interface

    def __delitem__(self, index):
        del self.points[index]

    def __getitem__(self, index):
        return self.points[index]

    def __setitem__(self, index, point):
        if not isinstance(point, Point):
            raise TypeError("expected Point, found %s" % type(point).__name__)
        self.points[index] = point

    def __iter__(self):
        return iter(self.points)

    def __len__(self):
        return len(self.points)

    def insert(self, index, point):
        if not isinstance(point, Point):
            raise TypeError("expected Point, found %s" % type(point).__name__)
        self.points.insert(index, point)

    # TODO: rotate method?

    @property
    def open(self):
        if not self.points:
            return True
        return self.points[0].type == "move"

    # -----------
    # Pen methods
    # -----------

    def draw(self, pen):
        pointPen = PointToSegmentPen(pen)
        self.drawPoints(pointPen)

    def drawPoints(self, pointPen):
        try:
            pointPen.beginPath(identifier=self.identifier)
            for p in self.points:
                pointPen.addPoint(
                    (p.x, p.y),
                    segmentType=p.type,
                    smooth=p.smooth,
                    name=p.name,
                    identifier=p.identifier,
                )
        except TypeError:
            pointPen.beginPath()
            for p in self.points:
                pointPen.addPoint(
                    (p.x, p.y),
                    segmentType=p.type,
                    smooth=p.smooth,
                    name=p.name,
                )
            warnings.warn(
                "The pointPen needs an identifier kwarg. "
                "Identifiers have been discarded.",
                UserWarning,
            )
        pointPen.endPath()



import attr
import os
import shutil
import fs.tempfs
from ufoLib2.constants import DEFAULT_LAYER_NAME
from ufoLib2.objects.dataSet import DataSet
from ufoLib2.objects.guideline import Guideline
from ufoLib2.objects.imageSet import ImageSet
from ufoLib2.objects.info import Info
from ufoLib2.objects.layerSet import LayerSet
from ufoLib2.objects.features import Features
from fontTools.misc.py23 import basestring, PY3
from fontTools.ufoLib import UFOReader, UFOWriter, UFOFileStructure


@attr.s(slots=True, kw_only=PY3, repr=False)
class Font(object):
    layers = attr.ib(
        default=attr.Factory(LayerSet),
        validator=attr.validators.instance_of(LayerSet),
        type=LayerSet,
    )
    info = attr.ib(
        default=attr.Factory(Info),
        converter=lambda v: v if isinstance(v, Info) else Info(**v),
        type=Info,
    )
    features = attr.ib(
        default=attr.Factory(Features),
        converter=lambda v: v if isinstance(v, Features) else Features(v),
        type=Features,
    )
    groups = attr.ib(default=attr.Factory(dict), type=dict)
    kerning = attr.ib(default=attr.Factory(dict), type=dict)
    lib = attr.ib(default=attr.Factory(dict), type=dict)
    data = attr.ib(
        default=attr.Factory(DataSet),
        converter=lambda v: v if isinstance(v, DataSet) else DataSet(**v),
        type=DataSet,
    )
    images = attr.ib(
        default=attr.Factory(ImageSet),
        converter=lambda v: v if isinstance(v, ImageSet) else ImageSet(**v),
        type=ImageSet,
    )

    _path = attr.ib(default=None, init=False)
    _reader = attr.ib(default=None, init=False)
    _fileStructure = attr.ib(default=None, init=False)

    @classmethod
    def open(cls, path, lazy=True, validate=True):
        reader = UFOReader(path, validate=validate)
        self = cls.read(reader, lazy=lazy)
        self._path = path
        self._fileStructure = reader.fileStructure
        if lazy:
            # keep the reader around so we can close it when done
            self._reader = reader
        else:
            reader.close()
        return self

    @classmethod
    def read(cls, reader, lazy=True):
        layers = LayerSet.read(reader, lazy=lazy)
        data = DataSet.read(reader, lazy=lazy)
        images = ImageSet.read(reader, lazy=lazy)
        info = Info()
        reader.readInfo(info)
        features = Features(reader.readFeatures())
        groups = reader.readGroups()
        kerning = reader.readKerning()
        lib = reader.readLib()
        self = cls(
            layers=layers,
            info=info,
            features=features,
            groups=groups,
            kerning=kerning,
            lib=lib,
            data=data,
            images=images,
        )
        return self

    def __contains__(self, name):
        return name in self.layers.defaultLayer

    def __delitem__(self, name):
        del self.layers.defaultLayer[name]

    def __getitem__(self, name):
        return self.layers.defaultLayer[name]

    def __setitem__(self, name, glyph):
        self.layers.defaultLayer[name] = glyph

    def __iter__(self):
        return iter(self.layers.defaultLayer)

    def __len__(self):
        return len(self.layers.defaultLayer)

    def get(self, name, default=None):
        return self.layers.defaultLayer.get(name, default)

    def keys(self):
        return self.layers.defaultLayer.keys()

    def close(self):
        if self._reader is not None:
            self._reader.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()

    def __repr__(self):
        names = list(filter(None, [self.info.familyName, self.info.styleName]))
        fontName = " '{}'".format(" ".join(names)) if names else ""
        return "<{}.{}{} at {}>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            fontName,
            hex(id(self)),
        )

    @property
    def reader(self):
        return self._reader

    @property
    def glyphOrder(self):
        return list(self.lib.get("public.glyphOrder", []))

    @glyphOrder.setter
    def glyphOrder(self, value):
        if value is None or len(value) == 0:
            value = 0
            if "public.glyphOrder" in self.lib:
                del self.lib["public.glyphOrder"]
        else:
            self.lib["public.glyphOrder"] = value

    @property
    def guidelines(self):
        if self.info.guidelines is None:
            return []
        return self.info.guidelines

    @guidelines.setter
    def guidelines(self, value):
        for guideline in value:
            self.appendGuideline(guideline)

    @property
    def path(self):
        try:
            return self._path
        except AttributeError:
            return

    def addGlyph(self, glyph):
        self.layers.defaultLayer.addGlyph(glyph)

    def newGlyph(self, name):
        return self.layers.defaultLayer.newGlyph(name)

    def newLayer(self, name, **kwargs):
        return self.layers.newLayer(name, **kwargs)

    def renameGlyph(self, name, newName, overwrite=False):
        self.layers.defaultLayer.renameGlyph(name, newName, overwrite)

    def renameLayer(self, name, newName, overwrite=False):
        self.layers.renameLayer(name, newName, overwrite)

    def appendGuideline(self, guideline):
        if not isinstance(guideline, Guideline):
            guideline = Guideline(**guideline)
        if self.info.guidelines is None:
            self.info.guidelines = []
        self.info.guidelines.append(guideline)

    def write(self, writer, saveAs=None):
        if saveAs is None:
            saveAs = self._reader is not writer
        # TODO move this check to fontTools UFOWriter
        if self.layers.defaultLayer.name != DEFAULT_LAYER_NAME:
            assert DEFAULT_LAYER_NAME not in self.layers.layerOrder
        # save font attrs
        writer.writeFeatures(self.features.text)
        writer.writeGroups(self.groups)
        writer.writeInfo(self.info)
        writer.writeKerning(self.kerning)
        writer.writeLib(self.lib)
        # save the layers
        self.layers.write(writer, saveAs=saveAs)
        # save bin parts
        self.data.write(writer, saveAs=saveAs)
        self.images.write(writer, saveAs=saveAs)

    def save(
        self,
        path=None,
        formatVersion=3,
        structure=None,
        overwrite=False,
        validate=True,
    ):
        if formatVersion != 3:
            raise NotImplementedError(
                "unsupported format version: %s" % formatVersion
            )
        # validate 'structure' argument
        if structure is not None:
            structure = UFOFileStructure(structure)
        elif self._fileStructure is not None:
            # if structure is None, fall back to the same as when first loaded
            structure = self._fileStructure

        if hasattr(path, "__fspath__"):
            path = path.__fspath__()
        if isinstance(path, basestring):
            path = os.path.normpath(path)
        # else we assume it's an fs.BaseFS and we pass it on to UFOWriter

        overwritePath = tmp = None

        saveAs = path is not None
        if saveAs:
            if isinstance(path, basestring) and os.path.exists(path):
                if overwrite:
                    overwritePath = path
                    tmp = fs.tempfs.TempFS()
                    path = tmp.getsyspath(os.path.basename(path))
                else:
                    import errno

                    raise OSError(
                        errno.EEXIST, "path %r already exists" % path
                    )
        elif self.path is None:
            raise TypeError("'path' is required when saving a new Font")
        else:
            path = self.path

        try:
            with UFOWriter(
                path, structure=structure, validate=validate
            ) as writer:
                self.write(writer, saveAs=saveAs)
            writer.setModificationTime()
        except Exception:
            raise
        else:
            if overwritePath is not None:
                # remove existing then move file to destination
                if os.path.isdir(overwritePath):
                    shutil.rmtree(overwritePath)
                elif os.path.isfile(overwritePath):
                    os.remove(overwritePath)
                shutil.move(path, overwritePath)
                path = overwritePath
        finally:
            # clean up the temporary directory
            if tmp is not None:
                tmp.close()

        self._path = path
